/***************************************************************************** 
 * File        : mm.c
 * Date        : 2007/04/04
 * Author      : Sun Joon Gyu
 * Modified    : 2007/04/04
 * Description : Rutines for memory management
 * ***************************************************************************/

#include"mm.h"
#include"../thread/thread.h"
#include"../arch/port.h"


extern tcb_t *pCurrThread_g;
extern tcb_t *pNextThread_g;

unsigned int *pHeapList_g;	/* a head for heap spaces */
unsigned int *pHeapTail_g;	/* It indicates the end of heap */

char heap[1024*1024];		/* only for testing this rutines */




/******************************************************************
 * Function : Kern_MemInit()
 * Date	: 2007/04/10
 * Written by : Sun Joon Gyu
 * Operation : It simply initialize memory control block for heap.
 *             At first, it craetes a header controlling memory spaces.
 *             And it puts the header into the top of the heap.
 *****************************************************************/

char Kern_MemInit()
{
	
	pHeapList_g = (unsigned int *)START_HEAP_ADDR;
	pHeapTail_g = NULL;
	
	/* Initializing a header of heap*/
	memHead_t memHeader;
	memHeader.pPrev = (pMemHead_t) pHeapList_g;
	memHeader.pNext = (pMemHead_t) pHeapTail_g;
	memHeader.status = AVAILABLE;
	memHeader.size = (unsigned int) (((char *)END_HEAP_ADDR - (char *)pHeapList_g ) / BLOCK_SIZE) - BLOCK_SIZE; /* should subtract the size of one memory block */

	if(memHeader.size <= 0)
		return FAILED;

	/*Put the header just finished initializing into first location of heap*/
	*((pMemHead_t)pHeapList_g) = memHeader;

	return SUCCESS;

}



/***********************************************************************************************
 * Function : Kern_WantedBlkFind(size_t size)
 * Date	: 2007/04/10
 * Written by : Sun Joon Gyu
 * Arguments : size_t size - the number of bytes desired
 * Return type : pMemHead_t - it returns an address indicating specific memory which is available.
 * Operation : Sequencially, It goes around memory control blocks until it finds a memory control 
 *             block which can be allowed to put desired size of bytes
 ***********************************************************************************************/


pMemHead_t Kern_WantedBlkFind(size_t size)
{
CRITICAL_SECTION_ENTER();
	pMemHead_t pBlk = (pMemHead_t)pHeapList_g;

	size += (BLOCK_SIZE-1);
	size /= BLOCK_SIZE;

	for(;pBlk != NULL;pBlk = pBlk->pNext){
		if(pBlk->size >= size){
CRITICAL_SECTION_LEAVE();
			return pBlk;
		}
	}
CRITICAL_SECTION_LEAVE();

	return NULL;
		
}





/*************************************************************************************************
 * Function : Kern_MemBlkGet()
 * Date	: 2007/04/10
 * Written by : Sun Joon Gyu
 * Arguments : pMemHead_t pBlkAddr - the address picked in the function Kern_WantedBlkFind() 
 *             size_t size - the number of bytes desired
 * Return type : char - It returns a condition value indicating if this rutines completed successfully or not
 * Operation : It adjust neibor's links and create new memory control block to control extra memory spaces
 * ************************************************************************************************/


char Kern_MemBlkGet(pMemHead_t pBlkAddr,size_t size)
{

	size += (BLOCK_SIZE-1);
	size /= BLOCK_SIZE;
	
	printf("start \n");
CRITICAL_SECTION_ENTER();
	/* Just in case there is something wrong with memory block pointer */
	if((pBlkAddr->size) == size){

		pBlkAddr->status = ALLOCATED;
		pBlkAddr->pPrev->pNext = pBlkAddr->pNext;
		pBlkAddr->pNext->pPrev = pBlkAddr->pPrev;
CRITICAL_SECTION_LEAVE();
	}
	else if((pBlkAddr->size) > size){
		printf("bigger \n");
		pBlkAddr->status = ALLOCATED;
CRITICAL_SECTION_LEAVE();
		memHead_t memHeader;
		size_t extraBlkSize = pBlkAddr->size - size - 1;  
			/* should subtract an additional memory control header */

		memHeader.size = extraBlkSize;
		memHeader.pPrev = pBlkAddr->pPrev;
		memHeader.pNext = pBlkAddr->pNext;
		memHeader.status = AVAILABLE;

		
		pMemHead_t pExtraMemBlk = (pMemHead_t)((char *)pBlkAddr
				+ ((size + 1) * BLOCK_SIZE)); 
CRITICAL_SECTION_ENTER();
		pHeapList_g = (unsigned int *)pExtraMemBlk;

		*((pMemHead_t)pExtraMemBlk) = memHeader;

		pBlkAddr->pPrev->pNext = pExtraMemBlk;
		pBlkAddr->pNext->pPrev = pExtraMemBlk;
		pBlkAddr->pNext = pExtraMemBlk;
		pBlkAddr->size = size;
		pBlkAddr->status = ALLOCATED;
CRITICAL_SECTION_LEAVE();
	}
	else{

CRITICAL_SECTION_LEAVE();
		return FAILED;
	}


	return SUCCESS;	

}	






/*****************************************************************************
 * Function : Kern_BlkTblRegister()
 * Date	: 2007/04/10
 * Written by : Sun Joon Gyu
 * Arguments : unsigned int *pReqVar - a pointer variable which is asking the 
 * 					bytes of memory spaces
 *             pMemHead_t - A address pointing at the location which is just 
 *             		    allocated
 *
 * Return type : char - It returns a condition value indicating if this rutines 
 * 			completed successfully of not
 * Operation : It registers a address of a desired pointer variable into memory 
 * 		control block and also put the control block into the list in 
 * 		thread control block
 ******************************************************************************/

char Kern_BlkTblRegister(unsigned int *pReqVar, pMemHead_t pBlkAddr)
{
CRITICAL_SECTION_ENTER();
	if(pCurrThread_g!= NULL){
		if(pCurrThread_g->pMemList != NULL){
			pBlkAddr->pAllocNext = pCurrThread_g->pMemList;
			pCurrThread_g->pMemList = pBlkAddr;
		}else{
			pBlkAddr->pAllocNext = NULL;
			pCurrThread_g->pMemList = pBlkAddr;

		}
	}

	pBlkAddr->pVar = pReqVar;
CRITICAL_SECTION_LEAVE();
}





/*************************************************************************************************
 * Function :  MemRelease()
 * Date	: 2007/04/10
 * Written by : Sun Joon Gyu
 * Arguments : void *p - the address of memory wanted to be released
 * Return type : char - It returns a condition value indicating if this rutines completed successfully of not
 * Operation : it simply makes the bunch of memory blocks owned by the thread free. 
 ***************************************************************************************************/

char MemRelease(void *p)
{

	pMemHead_t pIter = pCurrThread_g->pMemList;

	
	for(; pIter != NULL; pIter = pIter->pAllocNext){
		if(pIter == p){
			if(pIter->pAllocPrev == NULL){
				if(pIter->pAllocNext == NULL){
					pCurrThread_g->pMemList = NULL;
				} else {
					pIter->pAllocNext->pPrev = (pMemHead_t)0;
					pCurrThread_g->pMemList = pIter->pAllocNext;
				}
			} else {
				if(pIter->pAllocNext == NULL){
					pIter->pAllocPrev->pAllocNext = (pMemHead_t)NULL;
				} else {
					pIter->pAllocPrev->pAllocNext = pIter->pAllocNext;
					pIter->pAllocNext->pAllocPrev = pIter->pAllocPrev;
				}
			}
				
		}
	}

/*CRITICAL_SECTION_ENTER() */
	if(((pMemHead_t)((char *)p - BLOCK_SIZE))->pPrev->status != ALLOCATED){
		pMemHead_t pPrevBlk = ((pMemHead_t)((char *)p-BLOCK_SIZE))->pPrev;
		pMemHead_t pCurrBlk = ((pMemHead_t)((char *)p-BLOCK_SIZE));

		/* merging */
		Kern_MemMerge(pPrevBlk, pCurrBlk, BACKWARD);
		
	}


	if(((pMemHead_t)((char *)p - BLOCK_SIZE))->pNext->status != ALLOCATED){
		pMemHead_t pNextBlk = ((pMemHead_t)((char *)p-BLOCK_SIZE))->pNext;
		pMemHead_t pCurrBlk = ((pMemHead_t)((char *)p-BLOCK_SIZE));

		/* merging */
		Kern_MemMerge(pCurrBlk, pNextBlk, FORWARD);

		pHeapList_g = (unsigned int *)pCurrBlk;
	}
/*CRITICAL_SECTION_EXIT() */
			
			
		
	
}



/*************************************************************************************************
 * Function :  Kern_MemMerge()
 * Date	: 2007/04/10
 * Written by : Sun Joon Gyu
 * Arguments : pMemHead_t pPrevBlk - 
 * Return type : char - It returns a condition value indicating if this rutines completed successfully of not
 * Operation : it simply makes the bunch of memory blocks owned by the thread free. 
 ***************************************************************************************************/

char Kern_MemMerge(pMemHead_t pPrevBlk, pMemHead_t pNextBlk,unsigned char dir)
{

	unsigned int mergedSize = (pPrevBlk->size + pNextBlk->size) + 1;	/* include one block used as a header of memory blocks */

	if(dir == BACKWARD){
		pNextBlk->pNext->pPrev= pPrevBlk;
		pPrevBlk->pNext = pNextBlk->pNext;
		pPrevBlk->size = mergedSize;
	}
	else if(dir = FORWARD){

		pNextBlk->pPrev->pNext = pPrevBlk;	/* Now we need to adjust other available blocks' links around this */
		pNextBlk->pNext->pPrev = pPrevBlk;      /* because the header of the available memory control block	 */
					   	        /* will be moved to previous memory control block */

		pPrevBlk->pNext = pNextBlk->pNext; 
		pPrevBlk->pPrev = pNextBlk->pPrev;
		pPrevBlk->size = mergedSize;
		pPrevBlk->status = AVAILABLE;
		pPrevBlk->threadID = 0;	
	}

	return SUCCESS;	

}

unsigned int *MemAlloc(unsigned int *pReqVar, size_t size)
{

	unsigned char err;
	pMemHead_t pBlkAddr;

	if( (pBlkAddr = Kern_WantedBlkFind(size)) == NULL){
		/* The thread tried to get memory from heap area should be blocked!! */
		/* DbgPrint("Not enough memory in Heap area\n"); */
	}
	if(Kern_MemBlkGet(pBlkAddr, size) == FAILED){
		/* DbgPrint("Something wrong!!\n"); */
	}

	if((Kern_BlkTblRegister(pReqVar, pBlkAddr)) ){
		/* The block requested to regist on block management table for compaction is already allocated!! */
		/* DbgPrint("The block requested is already allocated before!!\n"); */
	}

	return (unsigned int *)((char *)pBlkAddr + BLOCK_SIZE);	/* return the address except memory header */
}

/*
int main()
{
	int k = Kern_MemInit();
	char *p;
	unsigned int *addr;
	int i;
	addr = pHeapList_g;
	printf("%d",k);
	printf("%p", ((unsigned int *)addr));

	printf("\n%d\n",((pMemHead_t)pHeapList_g)->size);		
	p = MemAlloc((unsigned int *)&p,61);
	printf("%p\n",p);	
	for(i = 0; i < 50; i++){
		p[i] = 'i';
	}
	p[i] = '\0';
	printf("%s",p);	
	MemRelease(p);
	printf("\n");
	printf("%d",((pMemHead_t)pHeapList_g)->size);		
	printf("\n");

	MemAlloc((unsigned int *) &p, 40);
	printf("%p\n",	p);
	printf("%p\n", ((unsigned int *)addr));
	printf("%d\n",((pMemHead_t)pHeapList_g)->size);		
	printf("%p\n", ((pMemHead_t)addr)->pPrev);
}*/
